#include <glylib.h>
#include <AMBER/amber.h>


#define USAGE "\n\
\trun_sugarid input.pdb \n"


/*<<<defining various structures for use in the program>>>*/ 


typedef struct {
        int n; /**< number of atoms */
        char **N;
} natoms;

typedef struct{
	char *O;
	char *C;
	char *R;
	char *Ra;
	char *RN;
	int Rn;
} cxns;

typedef struct{
        char *sugar_type;
        char sugar_code[100];
        char *deoxy;
}sugarname;

typedef struct{
        int n;
        char *link_atom;

}links;
//aatoms=anomeric atoms;

/*<<<subfunctions used in the program>>>*/

void extract_models(fileslurp Fslurp);
void clean_protein(assembly *P);
void process_protein(char* filename);
molecule process_sugar(char* filename);
natoms* detect_cycle(residue S);
cxns* detect_anomer(natoms*,residue S,molecule M);
natoms* sort_atoms(natoms*,cxns*,coord_3D *cycle_coords);
sugarname* identify_name(residue S,int resno,natoms*,cxns*,molecule M);
char* assign_name(sugarname*,residue S,molecule M,natoms*);
links*  identify_linkage(residue S,molecule M,cxns*,natoms*);



int main(int argc,char *argv[]){
	

if(argc<2){mywhineusage("Insufficient number of arguments on command line.",USAGE);}
if(argc>2){mywhineusage("Too many arguments on command line.",USAGE);}

fileset Iset;
fileslurp Islurp;
int i=0;

/*<<<opening the input PDB file>>>*/
Iset.N=strdup(argv[1]);
Islurp=slurp_file(Iset);

/*<<<creating a PDB file with the protein and sugar, look into the function for more details>>>*/
extract_models(Islurp);

//printf("done\n");
fileset IRset,IRPset,IRSset;
fileslurp IRslurp;
IRPset.N=strdup("inter_protein.pdb");
IRSset.N=strdup("inter_sugar.pdb");
IRPset.F=myfopen(IRPset.N,"w");
IRSset.F=myfopen(IRSset.N,"w");
IRset.N=strdup("all.pdb");
IRslurp=slurp_file(IRset);
//printf("done\n");

/*<<<scanning the all.pdb file and writing all the ATOM cards to a file inter_protein.pdb and HETATM cards to a file inter_sugar.pdb>>>*/
for(i=0;i<IRslurp.n;i++){
	if(strncmp(IRslurp.L[i],"ATOM",4) == 0 || strncmp(IRslurp.L[i],"TER",3) == 0){
		fprintf(IRPset.F,"%s",IRslurp.L[i]);
	}
	else{
	if(strncmp(IRslurp.L[i],"HETATM",6)== 0 ||strncmp(IRslurp.L[i],"TER",3) == 0){
		if(strstr(IRslurp.L[i],"HOH") == NULL){
		fprintf(IRSset.F,"%s",IRslurp.L[i]);
		}
	}
	}
}
fclose(IRSset.F);
fclose(IRPset.F);
//printf("done\n");

char *sugarfile;
sugarfile=(char*)calloc(10,sizeof(char));
sprintf(sugarfile,"sugar.pdb");
//Sset.N=strdup("sugar.pdb");


/*<<<output file for writing the output>>>*/
fileset Oset;
Oset.N=strdup("test_output");
Oset.F=myfopen(Oset.N,"w");
int mi=0,ri=0,j=0;
molecule Sugar,temp_sugar;
Sugar.nr=0;
Sugar.r=(residue*)calloc(0,sizeof(residue));
assembly *B;
char IRSfile[16];
sprintf(IRSfile,"inter_sugar.pdb");
/*<<<processing the inter_sugar.pdb (looking for residues that do not have atleast 5 atoms and eliminating then etc. Pls look into the function for more details>>>*/
temp_sugar=process_sugar(IRSfile);

char tempsugarfile[15];
sprintf(tempsugarfile,"temp_sugar.pdb");
outputMolPDB(&temp_sugar,tempsugarfile);

//printf("created inter_sugar_file\n");
B=load_pdb(tempsugarfile);
printf("<<<<<<<The number of molecules are %d>>>>>>>\n",B[0].nm);
		

/*<<<copying from the inter_sugar to molecule Sugar, i think i did this because if there are sugars in different molecules they can all be combined into one molecule Sugar>>>*/
for(mi=0;mi<B[0].nm;mi++){
	for(ri=0;ri<B[0].m[mi][0].nr;ri++){
			Sugar.nr++;
			Sugar.r=(residue*)realloc(Sugar.r,(Sugar.nr*sizeof(residue)));
			Sugar.r[Sugar.nr-1]=B[0].m[mi][0].r[ri];
	}
			
			
}


/*<<<checking if the number of sugar residues are less than 0>>>*/
if(Sugar.nr<1){
	//printf("SORRY NO SUGAR RESIDUES FOUND IN THIS FILE!!!!\n");
        //printf("Please check your input file!\n");
        exit(1);
}

molecule S;
if(Sugar.nr>=1){
	 //printf("<<<<<<< There are %d carbohydrate residues >>>>>>>\n",Sugar.nr);
	outputMolPDB(&Sugar,sugarfile); /*<<<creating a file sugar.pdb>>>*/ 
	S=process_sugar(sugarfile);/*<<<preprocessing the sugar file to make sure there are no other residues>>>*/
}

char *processedfile=NULL;
processedfile=(char*)calloc(23,sizeof(char));
sprintf(processedfile,"sugar_preprocessed.pdb");
//printf("here\n");
/*<<<creating a file called sugar_preprocessed.pdb>>>*/
outputMolPDB(&S,processedfile);

//printf("created the sugar_preprocessed file\n");

/*<<<At this point the program has created a file called sugar_preprocessed.pdb which contains all the sugar residues. It is already loaded in the molecule S. Now, the program will try to identify each residue in this molecule>>>*/


/*<<<creating som arrays and structures for later use in the program>>>*/
i=0;j=0;
int csize=0,acsize=0,ssize=0,k=0,l=0,m=0,n=0,o=0,acheck=0,ccheck=0,check=0,osize=0,ocheck=0,scheck=0;
natoms* new;
natoms* second_cycle;
natoms* cycle_sorted;
sugarname* sname;
cxns* anomeric;
links* linkcards;
coord_3D* cycle_coords;
cycle_coords=(coord_3D*)calloc(0,sizeof(coord_3D));
char **other_atoms;
other_atoms=(char**)calloc(0,sizeof(char*));
vectormag_3D distances;
initialize_vectormag_3D(&distances);
residue R,N;
R.na=0;
R.a=(atom*)calloc(0,sizeof(atom));
N.na=0;
N.a=(atom*)calloc(0,sizeof(atom));
char **anomeric_carbons;
anomeric_carbons=(char**)calloc(0,sizeof(char*));
int *no_sorted;
no_sorted=(int*)calloc(0,sizeof(int));
char **sorted_atoms;
sorted_atoms=(char**)calloc(0,sizeof(char*));
char *sugar_name;


/*<<<this is where it is trying to identify the sugar from the PDB file>>>*/
/*<<<looping through each residue in the molecule S>>>*/ 
printf("The number of residues are %d\n",S.nr);
for(i=0;i<S.nr;i++){
	//printf("inside for\n");
	new=detect_cycle(S.r[i]); /*<<<detect cycle detects the atoms in the ring, new is a data structure which can store the number of atoms in the ring in new.n and the atom names of each of these atoms in new.N[i]>>>*/

	//printf("The number of atoms in a cycle are %d\n",new[0].n);	
	if(new[0].n ==0){
                printf("I can't find a cyclic ring structure in this residue. Please check the log file for more information\n");
        }
        if(new[0].n>0){
                /*<<<if the program detects that there is a five or six membered ring, then it looks for coordinates of these atoms and stores them in an array called cycle_coords>>>*/
                for(j=0;j<new[0].n;j++){
                        for(k=0;k<S.r[i].na;k++){
                                //printf("%s and %s\n",new.N[j],S.r[i].a[k].N)>>>*/
                                if(strcmp(new[0].N[j],S.r[i].a[k].N) == 0){
                                        csize++;
                                        cycle_coords=(coord_3D*)realloc(cycle_coords,(csize*sizeof(coord_3D)));
                                        cycle_coords[csize-1]=S.r[i].a[k].x;
                                }
                        }
                }

		/*for(checking=0;checking<csize;checking++){
			dprint_coord_3D(&cycle_coords[checking]);
		}*/

		/*<<<storing the atoms that are connected to the ring atoms (except hydrogens) in the residue in an array called other_atoms>>>*/
                for(l=0;l<csize;l++){
                        for(m=0;m<S.r[i].na;m++){
                                check=0;
                                for(n=0;n<new[0].n;n++){
                                        if(strcmp(S.r[i].a[m].N,new[0].N[n]) == 0){
                                                check=1;
                                        }
                                }
                                if(check==0){
                                        distances=coord_to_vec(subtract_coord(cycle_coords[l],S.r[i].a[m].x));
                                        if(distances.d<1.9 && strncmp(S.r[i].a[m].N,"H",1) != 0){
                                                osize++;
                                                other_atoms=(char**)realloc(other_atoms,(osize*sizeof(char*)));
                                                other_atoms[osize-1]=strdup(S.r[i].a[m].N);
                                        }
                                }
                        }

                }
		
		/*for(checking=0;checking<osize;checking++){
                        dprint_coord_3D(&cycle_coords[checking]);
                }*/

		/*<<<stroing the remaining atoms in residue R>>>*/
		for(m=0;m<S.r[i].na;m++){
                        check=0;
                        ocheck=0;
                        for(n=0;n<new[0].n;n++){
                                if(strcmp(S.r[i].a[m].N,new[0].N[n]) == 0){
                                        check=1;
                                }
                        }
                        for(o=0;o<osize;o++){
                                if(strcmp(S.r[i].a[m].N,other_atoms[o]) == 0){
                                        ocheck=1;
                                }
                        }
                        if(check==0 && ocheck==0){
                                R.na++;
                                R.a=(atom*)realloc(R.a,(R.na*sizeof(atom)));
                                R.a[R.na-1]=S.r[i].a[m];
                        }
                }

                //printf("The remaining atoms are %d\n",R.na);
                /*<<<not exactly sure why i wrote this loop but it was something to do with residues connected to each other when there is a double ring>>>*/
		check=0;
                if(R.na>0){
                for(m=0;m<R.na;m++){
                        distances=coord_to_vec(subtract_coord(R.a[m].x,R.a[0].x));
                        if(distances.d<1.9 && distances.d!= 0.0){
                                check=1;
                        }
                        //printf("The remaining atoms are %s\n",R.a[m].N);
                }
                }

		/*<<<checking to see if there is a second cyclic ring in the residue>>>*/
		scheck=0;
		if(check==0){
                        if(R.na>=6){
                                for(m=1;m<R.na;m++){
                                        N.na++;
                                        N.a=(atom*)realloc(N.a,(N.na*sizeof(atom)));
                                        N.a[N.na-1]=R.a[m];
                                }
			scheck=1;
                        second_cycle=detect_cycle(N);
                        }
                }
                else{
                        if(R.na>=6){
				//printf("inside if\n");
				scheck=1;
                		second_cycle=detect_cycle(R);
                        }
                }


		/*<<<if there is a second cycle then the sugar will be named 2XX and the rest of the program after this loop will not be executed>>>*/
		if(scheck==1 && second_cycle[0].n>0){
                        sugar_name=strdup("2XX");
                        //printf("detected second cycle\n");
                        printf("%d\t%s\n",i+1,sugar_name);
                        fprintf(Oset.F,"%d\t%s\n",i+1,sugar_name);

                }

		
		/*<<<if there is no second cyclic ring then it will look for the anomeric oxygen and carbon>>>*/
		acheck=0;
                if(scheck==1 && second_cycle[0].n<1){
			acheck=1;
                        anomeric =detect_anomer(new,S.r[i],S);
			//printf("The anomeric oxygen and carbon are %s and %s\n",anomeric[0].O,anomeric[0].C);
		}
		else{
			if(scheck==0 ){
				acheck=1;
				anomeric =detect_anomer(new,S.r[i],S);
                        	//printf("The anomeric oxygen and carbon are %s and %s\n",anomeric[0].O,anomeric[0].C);
			}
		}

		
		/*<<<if the anomeric carbon and oxygen are found then the program will try to identify the residue type and assign a residue name>>>*/ 
		ccheck=0;
		/*<<<if the anomeric carbon or oxygen is not detected then the program assigns the sugar name as 5XX and will not execute the remaining sub functions>>>*/
		if(acheck==1){
			if(strstr(anomeric[0].C,"NULL") != NULL){
                                //printf("inside anomeric\n");
                                acsize++;
                                anomeric_carbons=(char**)realloc(anomeric_carbons,(acsize*sizeof(char*)));
                                anomeric_carbons[acsize-1]=strdup("NULL");
				sugar_name=strdup("5XX");
				printf("%d\t%s\n",i+1,sugar_name);
				fprintf(Oset.F,"%d\t%s\n",i+1,sugar_name);
                        }

                        if(strstr(anomeric[0].C,"NULL") == NULL && strstr(anomeric[0].O,"NULL") == NULL){
                                acsize++;
                                anomeric_carbons=(char**)realloc(anomeric_carbons,(acsize*sizeof(char*)));
                                anomeric_carbons[acsize-1]=strdup(anomeric[0].C);
				//if tha anomeric carbon and oxygen are identified then the program will sort the atoms in the cyclic ring that it identified earlier
                                cycle_sorted= sort_atoms(new,anomeric,cycle_coords);
				ccheck=1;
                        }


			/*if anomeric carbon is found and the atoms in the cycle are sorted then it tries to identify the residue and assign a name*/
			if(cycle_sorted[0].n>0 && ccheck==1){
                                //printf("detected anomeric carbon and oxygen\n");
                                //printf("here\n");
                                no_sorted[i]=cycle_sorted[0].n;
                                for(l=0;l<cycle_sorted[0].n;l++){
                                        ssize++;
                                        sorted_atoms=(char**)realloc(sorted_atoms,(ssize*sizeof(char*)));
                                        sorted_atoms[ssize-1]=cycle_sorted[0].N[l];
                                }
                                //printf("sorted atoms in a cycle\n");
				//identifying the sugar type 
                                sname=identify_name(S.r[i],i,cycle_sorted,anomeric,S);
                                //printf("sugar_code is %s\n",sname[0].sugar_code);
                                //printf("sugar_type is %s\n",sname.sugar_type);
                                //printf("deoxy is %s\n",sname.deoxy);
				/*<<<assigning the name of the sugar>>>*/
                                sugar_name=assign_name(sname,S.r[i],S,cycle_sorted);
                                 //printf("sugarname is %s\n",sugar_name);
                                //printf("%d\t%s\n",i+1,sugar_name);
                                //fprintf(Oset.F,"%d\t%s\n",i+1,sugar_name);
				/*identifying the linkage*/
                                linkcards=identify_linkage(S.r[i],S,anomeric,new);
                        	printf("%d\t%s\t%s\t%d\t%s\n",i+1,sugar_name,anomeric[0].C,linkcards[0].n,linkcards[0].link_atom);
				fprintf(Oset.F,"%d\t%s\t%s\t%d\t%s\n",i+1,sugar_name,anomeric[0].C,linkcards[0].n,linkcards[0].link_atom);
			}


                        //printf("done\n");
		}
		


	}//if new.n>0


}//for



return 0;
}
